import os
import pwd
import argparse
import json
from threading import Thread, Event
from psycopg2 import OperationalError
from logging import getLogger
from sys import exit
from pg_logidater.exceptions import (
    PsqlConnectionError,
)
from pg_logidater.utils import (
    SqlConn,
    ServerConn,
    setup_logging,
    prepare_directories
)
from pg_logidater.master import (
    master_prepare,
    master_checks
)
from pg_logidater.replica import (
    pause_replica,
    replica_info,
)
from pg_logidater.tartget import (
    create_subscriber,
    create_database,
    target_check,
    sync_roles,
    sync_database,
    get_replica_position,
    sync_seq_pipe,
    analyse_target,
    db_sync_progress_bar
)


_logger = getLogger(__name__)
parser = argparse.ArgumentParser()
parser.add_argument(
    "--save-log",
    help="Save pg-logidater log to file",
    default="/tmp/pg-logidater.log"
)
parser.add_argument(
    "--app-tmp-dir",
    help="Temp directory to store dumps",
    type=str,
    default="/tmp/pg-logidater/tmp"
)
parser.add_argument(
    "--app-log-dir",
    help="Cli output log dir, default: /tmp/pg-logidater/log",
    type=str,
    default="/tmp/pg-logidater/log"
)
parser.add_argument(
    "--saved-conf",
    help="Path to file with saved json config",
    type=str
)
parser.add_argument(
    "-u",
    "--user",
    type=str,
    help="User for running application, default=postgres",
    default="postgres"
)
parser.add_argument(
    "--database",
    help="Database to setup logical replication",
)
parser.add_argument(
    "--master-host",
    help="Master host from which to setup replica",
    type=str,
)
parser.add_argument(
    "--replica-host",
    help="Replica host were to take dump",
    type=str,
)
parser.add_argument(
    "--psql-user",
    help="User for connecting to psql",
    type=str,
)
parser.add_argument(
    "--repl-name",
    help="Name for publication, subscription and replication slot",
    type=str,
)
log_level = parser.add_mutually_exclusive_group()
log_level.add_argument(
    "--log-level",
    type=str,
    choices=["debug", "info", "warning", "eror", "critical"],
    help="Log level for console outpu, default=info",
    default="info"
)
log_level.add_argument(
    "-d",
    "--debug",
    action="store_true"
)
log_level.add_argument(
    "--verbose",
    action="store_true"
)
subparser = parser.add_subparsers(dest="cli")


def argument(*name_of_flags, **kwargs) -> list:
    return (list(name_of_flags), kwargs)


def cli(args=[], parent=subparser, cmd_aliases=None):
    if cmd_aliases is None:
        cmd_aliases = []

        def decorator(func):
            parser = parent.add_parser(
                func.__name__.replace("_", "-"),
                description=func.__doc__,
                aliases=cmd_aliases
            )
            for arg in args:
                parser.add_argument(*arg[0], **arg[1])
            parser.set_defaults(func=func)
        return decorator


def drop_privileges(user) -> None:
    try:
        current_uid = os.getuid()
        change_user = pwd.getpwnam(user)
        if current_uid == change_user.pw_uid:
            return
        _logger.info(f"Chnaging user to: {user}")
        os.setgid(change_user.pw_gid)
        os.setuid(change_user.pw_uid)
        os.environ["HOME"] = change_user.pw_dir
    except PermissionError:
        _logger.error("Program must be executed as root!!")
        exit(1)
    except KeyError:
        _logger.error(f"{user} user doesn't exist")
        exit(1)


@cli(
    [
        argument(
            "--ignore-pkey",
            help="Ignore missing pkeys",
            action="store_true"
        ),
        argument(
            "--update-interval",
            help="Progress update interval, default 2s",
            default=2,
            type=float
        )
    ]
)
def setup_replica(args) -> None:
    try:
        master_sql = SqlConn(args["master_host"], user=args["psql_user"], db=args["database"])
        replica_sql = SqlConn(args["replica_host"], args["psql_user"])
        target_sql = SqlConn("/tmp", user="postgres", db="postgres")
    except PsqlConnectionError as e:
        _logger.critical(e)
    db_size = master_checks(
        psql=master_sql,
        slot_name=args["repl_name"],
        pub_name=args["repl_name"],
        skip_pkey_check=args["ignore_pkey"]
    )
    target_check(
        psql=target_sql,
        database=args["database"],
        name=args["repl_name"],
        db_size=db_size
    )
    with ServerConn(args["replica_host"], args["user"]) as ssh:
        app_name, slot_name = replica_info(
            psql=replica_sql,
            ssh=ssh
        )
    db_owner = master_prepare(
        psql=master_sql,
        name=args["repl_name"],
        database=args["database"]
    )
    create_database(
        psql=target_sql,
        database=args["database"],
        owner=db_owner
    )
    pause_replica(
        psql=replica_sql
    )
    replica_stop_position = get_replica_position(
        psql=master_sql,
        app_name=app_name
    )
    sync_roles(
        host=args["replica_host"],
        tmp_path=args["app_tmp_dir"],
        log_dir=args["app_log_dir"],
    )

    event_finished = Event()
    progress_thread = Thread(
        target=db_sync_progress_bar,
        args=(
            target_sql,
            db_size,
            event_finished,
            args["database"],
            args["update_interval"],
        )
    )
    progress_thread.start()
    sync_database(
        host=args["replica_host"],
        user=args["psql_user"],
        database=args["database"],
        tmp_dir=args["app_tmp_dir"],
        log_dir=args["app_log_dir"],
        event=event_finished
    )
    progress_thread.join()
    create_subscriber(
       sub_target=args["master_host"],
       database=args["database"],
       slot_name=args["repl_name"],
       repl_position=replica_stop_position
    )
    _logger.info("Rresuming replication")
    replica_sql.resume_replica()
    analyse_target(target_sql)


@cli()
def drop_setup(args) -> None:
    _logger.info("Cleaning target server")
    try:
        target_sql = SqlConn("/tmp", user="postgres", db=args["database"])
        target_sql.drop_subscriber(sub_name=args["repl_name"])
        target_sql = SqlConn("/tmp", user="postgres", db="postgres")
        target_sql.drop_database(args["database"])
    except OperationalError as err:
        _logger.warning(err)
    _logger.info("Cleaning up master")
    master_sql = SqlConn(args["master_host"], user=args["psql_user"], db=args["database"])
    master_sql.drop_publication(args["repl_name"])
    master_sql.drop_repl_slot(args["repl_name"])
    _logger.info("Cleaning up replica")
    replica_sql = SqlConn(args["replica_host"], args["psql_user"])
    replica_sql.resume_replica()


@cli()
def sync_sequences(args) -> None:
    master_sql = SqlConn(args["master_host"], user=args["psql_user"], db=args["database"])
    sync_seq_pipe(
        psql=master_sql,
        log_dir=args["app_log_dir"]
    )


@cli()
def remove_repl_config(args) -> None:
    _logger.info("Removing logical replication configuration")
    try:
        target_sql = SqlConn("/tmp", user="postgres", db=args["database"])
        _logger.debug(f"Dropping subscriber on localhost for db {args['database']}")
        target_sql.drop_subscriber(drop_slot=True, sub_name=args["repl_name"])
    except OperationalError as err:
        _logger.warning(err)
    master_sql = SqlConn(args["master_host"], user=args["psql_user"], db=args["database"])
    _logger.debug(f"Dropping publication on host {args['master_host']} for db {args['database']}")
    master_sql.drop_publication(args["repl_name"])


@cli(
    [
        argument(
            "--conf-save-path",
            help="Path were to save cli options for later use, default: current directory",
            default=os.path.join(os.getcwd(), "pg_logidater.conf")
        )
    ]
)
def save_cli_options(args: dict) -> None:
    _logger.info(f"Saving cli options to file {args['conf_save_path']}")
    with open(args["conf_save_path"], "w") as write_conf:
        args_dict = reduce_dict(args)
        args_dict.pop("func")
        json.dump(args_dict, write_conf, indent=2)


def reduce_dict(args: dict) -> dict:
    try:
        args.pop("conf_save_path")
    except KeyError:
        pass
    try:
        args.pop("cli")
    except KeyError:
        pass
    try:
        args.pop("saved_conf")
    except KeyError:
        pass
    return args


def resolve_config(args: dict) -> dict:
    _logger.info("Merging cli params and config file")
    conf_file = args["saved_conf"]
    conf = {}
    if os.path.isfile(conf_file):
        _logger.debug(f"Reading saved config from: {conf_file}")
        with open(conf_file, "r") as cf:
            conf = dict(json.load(cf))
            _logger.debug(f"Config file content:\n{json.dumps(conf, indent=2)}")
    else:
        raise FileNotFoundError(f"{conf_file} doesn't exists")
    reduce_dict(args)
    merged_dicts = args | conf
    return merged_dicts


def main():
    args = parser.parse_args()
    if args.cli is None:
        parser.print_help()
        exit(0)
    if args.debug:
        setup_logging(
            log_level="debug",
            debug_ssh=True,
            save_log=args.save_log,
            log_path=args.app_log_dir
        )
    elif args.verbose:
        setup_logging(
            log_level="debug",
            save_log=args.save_log,
            log_path=args.app_log_dir
        )
    else:
        setup_logging(
            log_level=args.log_level,
            save_log=args.save_log,
            log_path=args.app_log_dir
        )
    _logger.debug(f"Cli args: {args}")
    args_dict = vars(args)
    if args.cli == "save-cli-options":
        args.func(args_dict)
    else:
        if args.saved_conf is not None:
            args_dict = resolve_config(args_dict)
        drop_privileges(args_dict["user"])
        prepare_directories(args_dict["app_log_dir"], args_dict["app_tmp_dir"])
        args.func(args_dict)
        _logger.info(f"App debug log: {args.save_log}")
        _logger.info(f"Dump/restore logs: {args.app_log_dir}")


if __name__ == "__main__":
    main()



class PsqlConnectionError(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Unable to connect"
        super().__init__(message)


class DatabaseExists(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Database exists"
        super().__init__(message)


class ReplicaPaused(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Replica paused"
        super().__init__(message)


class ReplicaLevelNotCorrect(Exception):
    def __init__(self, message=None):
        if not message:
            message = "wal_level not correct"
        super().__init__(message)


class ReplicaSlotExists(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Replication slot exist"
        super().__init__(message)


class PublicationExists(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Publication exists"
        super().__init__(message)


class VersionNotSupported(Exception):
    def __init__(self, message=None):
        if not message:
            message = "Version not supported"
        super().__init__(message)


class DiskSpaceTooLow(Exception):
    """
    Disk space too low exception
    """



SQL_WAL_LEVEL = "SHOW wal_level"
SQL_IS_REPLICA_PASUSED = "SELECT pg_is_wal_replay_paused()"
SQL_PAUSE_REPLICA = "SELECT pg_wal_replay_pause()"
SQL_RESUME_REPLICA = "SELECT pg_wal_replay_resume()"
SQL_DROP_DATABASE = "DROP DATABASE {0}"
SQL_CREATE_DATABASE = "CREATE DATABASE {db} OWNER {owner}"
SQL_DROP_SUBSCRIPTION = "DROP SUBSCRIPTION {name}"
SQL_SHOW_VERSION = "SHOW server_version"
SQL_ADVANCE_REPLICA_POSITION = "SELECT pg_replication_origin_advance ('{id}', '{position}');"
SQL_ENABLE_SUBSCRIPTION = "ALTER SUBSCRIPTION {name} ENABLE"
SQL_SET_SLOT_NAME = "ALTER SUBSCRIPTION {name} SET (slot_name={slot_name})"
SQL_DISABLE_SUBSCRIPTION = "ALTER SUBSCRIPTION {name} DISABLE"
SQL_GET_DB_SUBSCRIPTION = "SELECT subname FROM pg_subscription"
SQL_CREATE_REPL_LOGICAL_SLOT = "SELECT pg_create_logical_replication_slot('{0}', 'pgoutput')"
SQL_DROP_REPL_SLOT = "SELECT pg_drop_replication_slot('{0}')"
SQL_CREAATE_PUB = "CREATE publication {0} for all tables"
SQL_DROP_PUB = "DROP publication {0}"
SQL_DATA_DIRECTORY = "SHOW data_directory"
SQL_DB_SIZE = "SELECT pg_database_size('{db}')"
SQL_ANALYZE = "ANALYZE VERBOSE"

SQL_SELECT_ALL_SEQUENCES = """
SELECT
  sequence_schema,
  sequence_name
FROM
  information_schema.sequences
"""

SQL_CREATE_SUBSCRIPTION = """
CREATE SUBSCRIPTION {name} connection 'host={master} port=5432 dbname={db} user=repmgr'
PUBLICATION {pub_name}
WITH
  (
    copy_data = FALSE,
    create_slot = FALSE,
    enabled = FALSE,
    slot_name = {repl_slot}
  )"""

SQL_CHECK_DATABASE = """
SELECT
  count(*)
FROM
  pg_database
WHERE
  datname = '{0}'"""

SQL_CHECK_REPLICA_SLOT = """
SELECT
  CASE
    WHEN active = FALSE
    AND slot_name = '{0}' THEN TRUE
    ELSE FALSE
  END AS slot_status
FROM
  pg_replication_slots
WHERE
  slot_name = '{0}'"""

SQL_CHECK_PUBLICATION = """
SELECT
  CASE WHEN
    pubname = '{0}'
      THEN true
      ELSE false
  END AS pub_status
FROM
  pg_publication
WHERE
  pubname = '{0}'"""

SQL_SELECT_DB_OWNER = """
SELECT
  pg_catalog.pg_get_userbyid (d.datdba)
FROM
  pg_catalog.pg_database d
WHERE
  d.datname = '{0}'"""

SQL_GET_REPLAY_LSN = """
SELECT
  replay_lsn
FROM
  pg_stat_replication
WHERE
  application_name = '{app_name}'"""

SQL_SELECT_SUB_NAME = """
SELECT
  roname
FROM
  pg_subscription sub,
  pg_replication_origin ro
WHERE
  'pg_' || sub.oid = ro.roname AND
  sub.subname = '{name}'"""

SQL_NO_PRIMARY_KEYS = """
SELECT
  t.table_schema,
  t.table_name
FROM
  information_schema.tables AS t
  LEFT JOIN information_schema.table_constraints AS tc ON (
    t.table_schema = t.table_schema
    AND t.table_name = tc.table_name
    AND tc.constraint_type = 'PRIMARY KEY'
  )
WHERE
  t.table_type = 'BASE TABLE'
  AND t.table_schema NOT IN ('pg_catalog', 'information_schema')
  AND tc.constraint_name IS NULL
ORDER BY
  t.table_schema"""



from logging import getLogger
from pg_logidater.utils import SqlConn, ServerConn
from os import path
from pg_logidater.exceptions import (
    ReplicaPaused,
    VersionNotSupported
)

_logger = getLogger(__name__)

RECOVYRY_CONF_BY_VERSION = {
    11: "recovery.conf",
    12: "postgresql.auto.conf"
}


def pause_replica(psql: SqlConn) -> None:
    _logger.info("Pausing replica")
    if psql.is_replica_pause():
        raise ReplicaPaused
    psql.pause_replica()


def replica_info(psql: SqlConn, ssh: ServerConn) -> (str, str):
    _logger.info("Collecting replica info")
    pgdata = psql.get_datadirectory()
    psql_version = int(psql.server_version())
    if psql_version not in RECOVYRY_CONF_BY_VERSION.keys():
        raise VersionNotSupported
    recovey_conf = RECOVYRY_CONF_BY_VERSION[psql_version]
    auto_conf_name = path.join(pgdata, recovey_conf)
    cli = f"cat {auto_conf_name}"
    _logger.debug(f"Executing: {cli}")
    psql_auto_conf = ssh.run_cmd(cli)
    for line in reversed(psql_auto_conf.splitlines()):
        if "application_name" in line:
            replica_app_name = line.split(" ")[-1].removeprefix("application_name=").strip("'")
            _logger.debug(f"Got replica app name: {replica_app_name}")
            break
    for line in reversed(psql_auto_conf.splitlines()):
        if "primary_slot_name" in line:
            replica_slot_name = line.split(" ")[-1].strip("'")
            _logger.debug(f"Got replica slot name: {replica_slot_name}")
            break
    return replica_app_name, replica_slot_name



from logging import getLogger
from pg_logidater.utils import SqlConn
from pg_logidater.exceptions import (
    PublicationExists,
    ReplicaLevelNotCorrect,
    ReplicaSlotExists,
)

PG_DUMP = "/usr/bin/pg_dump --no-publications --no-subscriptions"
PG_DUMPALL = "/usr/bin/pg_dumpall"
PSQL = "/usr/bin/psql"

_logger = getLogger(__name__)


def master_checks(psql: SqlConn, slot_name: str, pub_name: str, skip_pkey_check: bool) -> int:
    _logger.info("Executing master checks")
    _logger.debug("Checking wal_level")
    wal_level = psql.get_wal_level()
    if wal_level != "logical":
        raise ReplicaLevelNotCorrect(f"Current wal_level: {wal_level}, minimum required: logical")
    _logger.debug(f"Checking if replication slot {slot_name} doesn't exists")
    replica_slot = psql.get_replica_slot(slot_name)
    if replica_slot:
        raise ReplicaSlotExists(f"Replication slot {slot_name} already exists!")
    _logger.debug(f"Checking in publication {pub_name} doesn't exists")
    publication = psql.is_pub_exists(pub_name)
    if publication:
        raise PublicationExists(f"Publication {pub_name} already exists for db: {psql.sql_conn.get_dsn_parameters()['dbname']}")
    if not skip_pkey_check:
        _logger.debug("Checking tables without primary keys")
        no_keys_list = psql.get_no_primary_key()
        if no_keys_list:
            _logger.warning("Database have tables without primary key, that can cause issues with logical replication")
            _logger.warning("Following tables doesn't have primary key")
            for item in no_keys_list:
                print(f"{item[0]}.{item[1]}")
            _logger.warning("If that's fine, add --ignore-pkey switch")
            exit(1)
    _logger.debug("Getting db size")
    db = psql.sql_conn.get_dsn_parameters()["dbname"]
    db_size = psql.get_db_size(db)
    return db_size


def master_prepare(psql: SqlConn, name: str, database: str) -> str:
    _logger.info("Creating logical replication slot")
    psql.create_logical_slot(name)
    _logger.info("Creating publication")
    psql.create_publication(name)
    _logger.info(f"Geting db owner for {database}")
    db_owner = psql.get_database_owner(database)
    return db_owner



import logging
from sys import stdout
import paramiko
import psycopg2
import psycopg2.extras
from pg_logidater import sqlqueries as sql
from os import makedirs, path

LOG_FORMAT_CON = "[%(module)-8s:%(funcName)-20s| %(levelname)-8s] %(message)-40s"
LOG_FORMAT_FH = "[%(asctime)s - %(module)s:%(funcName)s|%(levelname)s] %(message)-40s"


_logger = logging.getLogger(__name__)


class ServerConn(paramiko.SSHClient):
    def __init__(self, host, user):
        super().__init__()
        self.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        self.load_system_host_keys()
        self.host = host
        self.user = user

    def __enter__(self):
        self.connect(hostname=self.host, username=self.user)
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def run_cmd(self, command: str) -> str:
        _, out, err = self.exec_command(command=command, timeout=60)
        error = err.read().decode()
        if len(error) > 0:
            _logger.error(f"Command {command} not found")
        return out.read().decode()


class SqlConn():
    def __init__(self, host, db="repmgr", user="repmgr", port="5432"):
        self.sql_conn = psycopg2.connect(
            host=host,
            port=port,
            user=user,
            database=db
        )
        self.cursor = self.sql_conn.cursor()
        if host.startswith("/tmp"):
            host = "localhost"
        _logger.debug(f"PSQL Connection to {host} with database {db} - established")

    def __del__(self) -> None:
        try:
            self.cursor.close()
            self.sql_conn.close()
        except AttributeError:
            pass

    def query(self, query, fetchone=False, fetchall=False) -> tuple:
        _logger.debug(f"Executing: {query}")
        try:
            self.cursor.execute(query)
        except psycopg2.errors.DuplicateObject as e:
            _logger.warning(f"Dublicate object {str(e).strip()}")
        self.sql_conn.commit()
        if fetchone:
            query_results = self.cursor.fetchone()
            return query_results
        if fetchall:
            query_results = self.cursor.fetchall()
            return query_results

    def create_logical_slot(self, slot_name):
        try:
            self.query(sql.SQL_CREATE_REPL_LOGICAL_SLOT.format(slot_name))
        except psycopg2.errors.DuplicateObject:
            _logger.warning(f"Replication slot {slot_name} exists")
            self.sql_conn.rollback()

    def drop_repl_slot(self, slot_name):
        try:
            self.query(sql.SQL_DROP_REPL_SLOT.format(slot_name))
        except psycopg2.errors.UndefinedObject:
            _logger.warning(f"Replication slot {slot_name} - doesn't exists")
            self.sql_conn.rollback()

    def create_publication(self, pub_name):
        try:
            self.query(sql.SQL_CREAATE_PUB.format(pub_name))
        except psycopg2.errors.DuplicateObject:
            _logger.warning(f"Publication {pub_name} - already exists")

    def drop_publication(self, pub_name):
        try:
            self.query(sql.SQL_DROP_PUB.format(pub_name))
        except psycopg2.errors.UndefinedObject:
            _logger.warning(f"Publication {pub_name} - doesn't exist")
            self.sql_conn.rollback()

    def create_subscriber(self, name, host, database, repl_slot) -> str:
        self.query(sql.SQL_CREATE_SUBSCRIPTION.format(name=name, master=host, db=database, pub_name=name, repl_slot=repl_slot))
        return self.query(sql.SQL_SELECT_SUB_NAME.format(name=name), fetchone=True)[0]

    def drop_subscriber(self, sub_name: str, drop_slot: bool = False) -> None:
        self.sql_conn.autocommit = True
        try:
            if sub_name:
                self.disable_subscription(sub_name)
                if not drop_slot:
                    self.set_slot_name(sub_name, "NONE")
                self.query(sql.SQL_DROP_SUBSCRIPTION.format(name=sub_name))
        except psycopg2.errors.UndefinedObject:
            _logger.warning(f"Subscription {sub_name} doesn't exist!")
        finally:
            self.sql_conn.autocommit = False

    def is_replica_pause(self) -> bool:
        if self.query(sql.SQL_IS_REPLICA_PASUSED, fetchone=True)[0]:
            _logger.warning("Replication is paused!")
            return True
        return False

    def pause_replica(self) -> None:
        self.query(sql.SQL_PAUSE_REPLICA)

    def resume_replica(self) -> None:
        self.query(sql.SQL_RESUME_REPLICA)

    def check_database(self, database) -> bool:
        if self.query(sql.SQL_CHECK_DATABASE.format(database), fetchone=True)[0] > 0:
            _logger.debug(f"Database {database} exists")
            return True
        return False

    def drop_database(self, database) -> None:
        _logger.debug(f"Droping database {database}")
        try:
            self.sql_conn.autocommit = True
            self.query(sql.SQL_DROP_DATABASE.format(database))
        except psycopg2.errors.ObjectInUse as err:
            _logger.error(err)
        finally:
            self.sql_conn.autocommit = False

    def create_database(self, database, owner="postgres") -> None:
        self.sql_conn.autocommit = True
        self.query(sql.SQL_CREATE_DATABASE.format(db=database, owner=owner))
        self.sql_conn.autocommit = False

    def get_database_owner(self, database) -> str:
        return self.query(sql.SQL_SELECT_DB_OWNER.format(database), fetchone=True)[0]

    def server_version(self) -> float:
        return float(self.query(sql.SQL_SHOW_VERSION, fetchone=True)[0])

    def get_replay_lsn(self, app_name) -> str:
        return self.query(sql.SQL_GET_REPLAY_LSN.format(app_name=app_name), fetchone=True)[0]

    def enable_subscription(self, sub_name, sub_id, pos) -> None:
        _logger.info(f"Changing starting position to {pos}")
        self.query(sql.SQL_ADVANCE_REPLICA_POSITION.format(id=sub_id, position=pos))
        _logger.info(f"Enabling subscription {sub_name}")
        self.query(sql.SQL_ENABLE_SUBSCRIPTION.format(name=sub_name))

    def disable_subscription(self, sub_name) -> None:
        _logger.info(f"Disabling subscription: {sub_name}")
        self.query(sql.SQL_DISABLE_SUBSCRIPTION.format(name=sub_name))

    def set_slot_name(self, sub_name, slot_name) -> None:
        _logger.info(f"Changing subscription {sub_name} slot to {slot_name}")
        self.query(sql.SQL_SET_SLOT_NAME.format(name=sub_name, slot_name=slot_name))

    def get_db_sub(self):
        _logger.info("Fetching subscription name")
        sub_name = self.query(sql.SQL_GET_DB_SUBSCRIPTION, fetchone=True)
        if isinstance(sub_name, tuple):
            _logger.debug(f"Subscription name: {sub_name[0]}")
            return sub_name[0]
        return False

    def get_wal_level(self) -> str:
        return self.query(sql.SQL_WAL_LEVEL, fetchone=True)[0]

    def get_replica_slot(self, name) -> str:
        slot = self.query(sql.SQL_CHECK_REPLICA_SLOT.format(name), fetchone=True)
        if isinstance(slot, tuple):
            return slot[0]

    def is_pub_exists(self, name) -> bool:
        if self.query(sql.SQL_CHECK_PUBLICATION.format(name), fetchone=True):
            return True
        return False

    def get_sequences(self) -> list[tuple]:
        return self.query(sql.SQL_SELECT_ALL_SEQUENCES, fetchall=True)

    def get_datadirectory(self) -> float:
        return (self.query(sql.SQL_DATA_DIRECTORY, fetchone=True))[0]

    def get_db_size(self, db) -> int:
        return (self.query(sql.SQL_DB_SIZE.format(db=db), fetchone=True))[0]

    def get_no_primary_key(self) -> list[tuple]:
        return self.query(sql.SQL_NO_PRIMARY_KEYS, fetchall=True)

    def analyze(self) -> None:
        self.query(sql.SQL_ANALYZE)


def setup_logging(log_level: str, save_log: str, debug_ssh: bool = False,  log_path: str = None) -> None:
    log_level_int = logging.getLevelName(str(log_level).upper())
    handlers = []
    if debug_ssh:
        logging.getLogger("paramiko").setLevel(logging.DEBUG)
    else:
        logging.getLogger("paramiko").setLevel(logging.WARNING)
    con_log = logging.StreamHandler(stdout)
    con_log.setFormatter(
        logging.Formatter(LOG_FORMAT_CON)
    )
    con_log.setLevel(log_level_int)
    handlers.append(con_log)
    if save_log:
        fh = logging.FileHandler(
            path.join(save_log)
        )
        fh.setFormatter(
            logging.Formatter(LOG_FORMAT_FH)
        )
        fh.setLevel(logging.DEBUG)
        handlers.append(fh)
    logging.basicConfig(
        level=logging.DEBUG,
        handlers=handlers
    )
    if save_log:
        _logger.info(f"Saving debug logging to {save_log}")


def prepare_directories(log_dir, tmp_dir) -> None:
    _logger.info("Creating needed directories")
    _logger.debug(f"Creating log dir: {log_dir}")
    makedirs(log_dir, exist_ok=True)
    _logger.debug(f"Creating tmp dir: {tmp_dir}")
    makedirs(tmp_dir, exist_ok=True)



from logging import getLogger
from pg_logidater.utils import SqlConn
from subprocess import Popen, PIPE
from os import path
from shutil import disk_usage
from pg_logidater.exceptions import (
    DatabaseExists,
    DiskSpaceTooLow
)
from threading import Event
from pycotore import ProgressBar

PG_DUMP_DB = "/usr/bin/pg_dump --no-publications --no-subscriptions -h {host} -U {user} {db}"
PG_DUMP_SEQ = "/usr/bin/pg_dump --no-publications --no-subscriptions -h {host} -d {db} -U {user} -t {seq_name}"
PG_DUMP_ROLES = "/usr/bin/pg_dumpall --roles-only -h {host} -U repmgr"
PSQL_SQL_RESTORE = "/usr/bin/psql -f {file} -d {db}"
PSQL_SQL_PIPE_RESTORE = "/usr/bin/psql -d {db}"


_logger = getLogger(__name__)


def target_check(psql: SqlConn, database: str, name: str, db_size: int) -> None:
    _logger.info("Executing target checks")
    _logger.debug("Checking available disk space")
    data_path = psql.get_datadirectory()
    available_disk = int(disk_usage(path=data_path).free * 0.9)
    if available_disk < db_size:
        raise DiskSpaceTooLow(f"Low disk space for {data_path}")
    _logger.debug("Checking if database not exists")
    if psql.check_database(database):
        raise DatabaseExists


def run_local_cli(cli, std_log, err_log, cli2: str = None, pipe: bool = False) -> None:
    with open(std_log, "w") as log:
        with open(err_log, "w") as err:
            if pipe:
                if not cli2:
                    _logger.critical("cli2 parameter mandaroty for pipe true")
                    exit(1)
                _logger.debug(f"Running: {cli} | {cli2}")
                pipe_output = Popen(cli.split(), stdout=PIPE)
                pipe_sync = Popen(cli2.split(), stdin=pipe_output.stdout, stdout=log, stderr=err)
                pipe_sync.communicate()
            else:
                _logger.debug(f"Executing: {cli}")
                Popen(cli.split(), stdout=log, stderr=err).communicate()


def get_replica_position(psql: SqlConn, app_name: str) -> str:
    _logger.info("Getting replication position")
    return psql.get_replay_lsn(app_name)


def sync_roles(host: str, tmp_path: str, log_dir: str) -> None:
    _logger.info("Syncing roles")
    roles_dump_path = path.join(tmp_path, "roles.sql")
    roles_dump_err_log = path.join(log_dir, "roles_dump.err")
    _logger.debug(f"Dumping roles to {roles_dump_path}")
    run_local_cli(
        PG_DUMP_ROLES.format(host=host),
        roles_dump_path,
        roles_dump_err_log
    )
    roles_restore_log = path.join(log_dir, "roles_restore.log")
    roles_restore_err_log = path.join(tmp_path, "roles_restore.err")
    _logger.debug(f"Restoring roles from {roles_dump_path}")
    run_local_cli(
        PSQL_SQL_RESTORE.format(file=roles_dump_path, db='postgres'),
        roles_restore_log,
        roles_restore_err_log
    )


def sync_database(host: str, user: str, database: str, tmp_dir: str, log_dir: str, event: Event) -> None:
    _logger.info(f"Syncing database {database}")
    sync_log = path.join(log_dir, f"sync_{database}.log")
    sync_err_log = path.join(log_dir, f"sync_{database}.err")
    event.set()
    run_local_cli(
        cli=PG_DUMP_DB.format(db=database, host=host, user=user),
        cli2=PSQL_SQL_PIPE_RESTORE.format(db=database),
        std_log=sync_log,
        err_log=sync_err_log,
        pipe=True
    )
    event.set()


def db_sync_progress_bar(psql: SqlConn, total: float, event: Event, db: str, update_interval: float) -> None:
    _logger.debug("Startign progress bar function")
    bar = ProgressBar()
    suffix = f"{db} sync in progress"
    bar.suffix = suffix
    bar.total = total
    event.wait(timeout=10)
    _logger.debug("Continue progrss function")
    event.clear()
    while True:
        if event.is_set():
            break
        synced_size = psql.get_db_size(db)
        bar.progress = synced_size
        bar.draw()
        event.wait(update_interval)
    bar.update_progress = total
    bar.draw()


def create_subscriber(sub_target: str, database: str, slot_name: str, repl_position: str) -> None:
    psql = SqlConn("/tmp", user="postgres", db=database)
    _logger.info(f"Creating subsriber to {sub_target}")
    sub_id = psql.create_subscriber(
        name=slot_name,
        host=sub_target,
        database=database,
        repl_slot=slot_name
    )
    psql.enable_subscription(
        sub_name=slot_name,
        sub_id=sub_id,
        pos=repl_position
    )


def create_database(psql: SqlConn, database: str, owner: str) -> None:
    _logger.info(f"Creating database {database}")
    psql.create_database(
        database=database,
        owner=owner
    )


def sync_seq_pipe(psql: SqlConn, log_dir: str) -> None:
    dsn = psql.sql_conn.get_dsn_parameters()
    database = dsn["dbname"]
    host = dsn["host"]
    user = dsn["user"]
    _logger.info(f"Syncing sequences for {database}")
    sequences = psql.get_sequences()
    for seq in sequences:
        file_seq_name = f"{seq[0]}.{seq[1]}"
        sql_seq_name = f"{seq[0]}.\"{seq[1]}\""
        sync_seq_log = path.join(log_dir, f"sync_seq_{file_seq_name}.log")
        sync_seq_err_log = path.join(log_dir, f"sync_seq_{file_seq_name}.err")
        run_local_cli(
            cli=PG_DUMP_SEQ.format(host=host, db=database, user=user, seq_name=sql_seq_name),
            cli2=PSQL_SQL_PIPE_RESTORE.format(db=database),
            std_log=sync_seq_log,
            err_log=sync_seq_err_log,
            pipe=True
        )


def analyse_target(psql: SqlConn) -> None:
    _logger.info("Updating database statistics")
    psql.analyze()
